"""
Configuration Loader
--------------------
Various functions that assist in loading initial configuration for the
control plane.

The control plane accepts a main configuration file from the environment
variable ``SOVEREIGN_CONFIG`` which follows the format:

.. code-block:: none

   <scheme>://path[,<scheme>://path,...]

Examples:

.. code-block:: none

   # Single file
     file:///etc/sovereign.yaml

   # Multiple files (comma separated)
     file:///etc/sovereign/common.yaml,file:///etc/sovereign/dev.yaml

   # Other types of sources
     http://config.myserver.com/environments/dev.yaml

   # Other types of formats
     http+json://config.myserver.com/environments/dev.json
     http+jinja://config.myserver.com/environments/dev.j2
     http+yaml://config.myserver.com/environments/dev.yaml

"""
import os
import json
import importlib
import yaml
import jinja2
import requests
from pkg_resources import resource_string

try:
    import ujson
except ImportError:
    import starlette.responses
    starlette.responses.ujson = json
    ujson = json

try:
    import boto3
except ImportError:
    boto3 = None


jinja_env = jinja2.Environment(enable_async=True)

serializers = {
    'yaml': yaml.safe_load,
    'json': json.loads,
    'ujson': ujson.loads,
    'jinja': jinja_env.from_string,
    'string': str,
}


def load_file(path, loader):
    with open(path) as f:
        contents = f.read()
        return serializers[loader](contents)


def load_package_data(path, loader):
    pkg, pkg_file = path.split(':')
    data = resource_string(pkg, pkg_file)
    try:
        data = data.decode()
    except AttributeError:
        pass
    return serializers[loader](data)


def load_http(path, loader):
    response = requests.get(path)
    response.raise_for_status()
    data = response.text
    return serializers[loader](data)


def load_env(variable, loader=None):
    data = os.getenv(variable)
    return serializers[loader](data)


def load_module(name, _=None):
    return importlib.import_module(name)


def load_s3(path: str, loader=None):
    if isinstance(boto3, type(None)):
        raise ImportError('boto3 must be installed to load S3 paths. Use ``pip install sovereign[boto]``')
    bucket, key = path.split('/', maxsplit=1)
    s3 = boto3.client('s3')
    response = s3.get_object(Bucket=bucket, Key=key)
    data = ''.join([chunk.decode() for chunk in response['Body']])
    return serializers[loader](data)


loaders = {
    'file': load_file,
    'pkgdata': load_package_data,
    'http': load_http,
    'https': load_http,
    'env': load_env,
    'module': load_module,
    's3': load_s3,
}


def parse_spec(spec, default_serialization='yaml'):
    serialization = default_serialization
    scheme, path = spec.split('://')
    if '+' in scheme:
        scheme, serialization = scheme.split('+')
    if 'http' in scheme:
        path = '://'.join([scheme, path])
    return scheme, path, serialization


def is_parseable(spec):
    if '://' not in spec:
        return False
    scheme, _, serialization = parse_spec(spec)
    return (
        scheme in loaders and
        serialization in serializers
    )


def load(spec):
    if '://' not in spec:
        return spec
    scheme, path, serialization = parse_spec(spec)
    return loaders[scheme](path, serialization)



import traceback
import uvicorn
from fastapi import FastAPI
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse
from sovereign import config, __versionstr__
from sovereign.sources import sources_refresh
from sovereign.views import crypto, discovery, healthchecks, admin, interface
from sovereign.middlewares import RequestContextLogMiddleware, get_request_id, LoggingMiddleware

try:
    import sentry_sdk
    from sentry_asgi import SentryMiddleware
except ImportError:
    sentry_sdk = None
    SentryMiddleware = None


def init_app() -> FastAPI:
    # Warm the sources once before starting
    sources_refresh()

    application = FastAPI(
        title='Sovereign',
        version=__versionstr__,
        debug=config.debug_enabled
    )
    application.include_router(discovery.router, tags=['Configuration Discovery'], prefix='/v2')
    application.include_router(crypto.router, tags=['Cryptographic Utilities'], prefix='/crypto')
    application.include_router(admin.router, tags=['Debugging Endpoints'], prefix='/admin')
    application.include_router(interface.router, tags=['User Interface'], prefix='/ui')
    application.include_router(healthchecks.router, tags=['Healthchecks'])

    application.add_middleware(RequestContextLogMiddleware)
    application.add_middleware(LoggingMiddleware)

    if config.sentry_dsn and sentry_sdk:
        sentry_sdk.init(config.sentry_dsn)
        # noinspection PyTypeChecker
        application.add_middleware(SentryMiddleware)

    @application.exception_handler(Exception)
    async def exception_handler(request: Request, exc: Exception):
        error = {
            'error': exc.__class__.__name__,
            'request_id': get_request_id()
        }

        # Add the description from Quart exception classes
        if hasattr(exc, 'detail'):
            error['description'] = getattr(exc.detail, 'description', 'unknown')

        if config.debug_enabled:
            error['traceback'] = [line for line in traceback.format_exc().split('\n')]
        status_code = getattr(exc, 'status_code', getattr(exc, 'code', 500))
        return JSONResponse(content=error, status_code=status_code)

    @application.get('/')
    def redirect_to_docs():
        return RedirectResponse('/docs')

    return application


app = init_app()


if __name__ == '__main__':
    uvicorn.run(app, host='0.0.0.0', port=8000, access_log=False)



import yaml
from collections import defaultdict
from fastapi import APIRouter, Query
from fastapi.encoders import jsonable_encoder
from starlette.responses import UJSONResponse
from sovereign.discovery import DiscoveryTypes
from sovereign.utils.mock import mock_discovery_request
from sovereign import discovery, config
from sovereign.sources import match_node
from sovereign.decorators import cache

router = APIRouter()


@router.get('/xds_dump')
async def display_config(
        xds_type: DiscoveryTypes = Query(..., title='xDS type', description='The type of request', example='clusters'),
        service_cluster: str = Query('*', title='The clients service cluster to emulate in this XDS request'),
        resource_names: str = Query('', title='Envoy Resource names to request'),
        region: str = Query(None, title='The clients region to emulate in this XDS request'),
        version: str = Query('1.11.1', title='The clients envoy version to emulate in this XDS request')
):
    ret = defaultdict(list)

    mock_request = mock_discovery_request(
        service_cluster=service_cluster,
        resource_names=resource_names,
        version=version,
        region=region
    )
    response = await discovery.response(
        request=mock_request,
        xds=xds_type.value,
        debug=True
    )
    if isinstance(response, dict):
        ret['resources'] += response.get('resources') or []

    return UJSONResponse(content=ret)


@router.get('/source_dump')
def instances(
        service_cluster: str = Query('*', title='The clients service cluster to emulate in this XDS request'),
        modified: str = Query('yes',
                              title='Whether the sources should run Modifiers/Global Modifiers prior to being returned')
):
    args = {
        'modify': yaml.safe_load(modified),
        'request': mock_discovery_request(
            service_cluster=service_cluster
        )
    }
    ret = match_node(**args)
    safe_response = jsonable_encoder(ret)
    return UJSONResponse(content=safe_response)


@router.get('/cache_dump')
def show_cached_keys():
    # noinspection PyProtectedMember
    return UJSONResponse(content=list(sorted(cache._cache.keys())))


@router.get('/config')
def show_configuration():
    return UJSONResponse(content=config.show())



import random
from fastapi.routing import APIRouter
from starlette.responses import PlainTextResponse
from sovereign import XDS_TEMPLATES, __versionstr__
from sovereign import discovery
from sovereign.sources import match_node
from sovereign.utils.mock import mock_discovery_request


router = APIRouter()


@router.get('/healthcheck', summary='Does the server respond')
async def health_check():
    return PlainTextResponse('OK')


@router.get('/deepcheck', summary='Can the server render a random template')
async def deep_check():
    template = random.choice(
        list(XDS_TEMPLATES['default'].keys())
    )
    await discovery.response(
        mock_discovery_request(),
        xds=template,
        debug=True
    )
    match_node(request=mock_discovery_request())
    return PlainTextResponse('OK')


@router.get('/version', summary='Display the current version of Sovereign')
async def version_check():
    return PlainTextResponse(f'Sovereign {__versionstr__}')



import schedule
from sovereign import config
from sovereign.utils import crypto
from sovereign.config_loader import load

template_context = {
    'crypto': crypto
}


def template_context_refresh():
    """ Modifies template_context in-place with new values """
    for k, v in config.template_context.items():
        template_context[k] = load(v)


# Initial setup
template_context_refresh()

if __name__ != '__main__' and config.refresh_context:
    schedule.every(config.context_refresh_rate).seconds.do(template_context_refresh)



import os
import zlib
from pydantic import BaseModel, Schema
from typing import List, Any
from jinja2 import Template
from sovereign.config_loader import load


class Source(BaseModel):
    type: str
    config: dict


class StatsdConfig(BaseModel):
    host: str = '127.0.0.1'
    port: int = 8125
    tags: dict = dict()
    namespace: str = 'sovereign'
    enabled: bool = False
    use_ms: bool = True

    @property
    def loaded_tags(self):
        return {k: load(v) for k, v in self.tags.items()}


class XdsTemplate(BaseModel):
    path: str

    @property
    def content(self) -> Template:
        return load(self.path)

    @property
    def checksum(self) -> int:
        return zlib.adler32(self.source.encode())

    @property
    def source(self) -> str:
        if 'jinja' in self.path:
            # The Jinja2 template serializer does not properly set a name
            # for the loaded template.
            # The repr for the template prints out as the memory address
            # This makes it really hard to generate a consistent version_info string
            # in rendered configuration.
            # For this reason, we re-load the template as a string instead, and create a checksum.
            path = self.path.replace('+jinja', '+string')
            return load(path)
        else:
            # The only other supported serializers are string, yaml, and json
            # So it should be safe to create this checksum off
            return str(self.content)


class Locality(BaseModel):
    region: str = Schema(None)
    zone: str = Schema(None)
    sub_zone: str = Schema(None)


class Node(BaseModel):
    id: str = Schema('-', title='Hostname')
    cluster: str = Schema(
        ...,
        title='Envoy service-cluster',
        description='The ``--service-cluster`` configured by the Envoy client'
    )
    build_version: str = Schema(
        ...,
        title='Envoy build/release version string',
        description='Used to identify what version of Envoy the '
                    'client is running, and what config to provide in response'
    )
    metadata: dict = Schema(None, title='Key:value metadata')
    locality: Locality = Schema(Locality(), title='Locality')


class Resources(list):
    """
    Acts like a regular list except it returns True
    for all membership tests when empty.
    """
    def __contains__(self, item):
        if len(self) == 0:
            return True
        return item in list(self)


class DiscoveryRequest(BaseModel):
    node: Node
    version_info: str = Schema('0', title='The version of the envoy clients current configuration')
    resource_names: Resources = Schema(Resources(), title='List of requested resource names')

    @property
    def envoy_version(self):
        try:
            build_version = self.node.build_version
            revision, version, *other_metadata = build_version.split('/')
        except (AttributeError, ValueError):
            # TODO: log/metric this?
            return 'default'
        return version

    @property
    def resources(self):
        return Resources(self.resource_names)


class DiscoveryResponse(BaseModel):
    version_info: str = Schema(..., title='The version of the configuration in the response')
    resources: List[Any] = Schema(..., title='The requested configuration resources')


class SovereignConfig(BaseModel):
    sources: List[Source]
    templates: dict
    template_context: dict = {}
    eds_priority_matrix: dict = {}
    modifiers: List[str] = []
    global_modifiers: List[str] = []
    regions: List[str] = []
    statsd: StatsdConfig = None
    auth_enabled: bool = os.getenv('SOVEREIGN_AUTH_ENABLED', False)
    auth_passwords: str = os.getenv('SOVEREIGN_AUTH_PASSWORDS', '')
    encryption_key: str = os.getenv('SOVEREIGN_ENCRYPTION_KEY', os.getenv('FERNET_ENCRYPTION_KEY'))
    no_changes_response_code: int = os.getenv('SOVEREIGN_NO_CHANGE_RESPONSE', 304)
    environment: str = os.getenv('SOVEREIGN_ENVIRONMENT_TYPE', os.getenv('MICROS_ENVTYPE', 'local'))
    debug_enabled: bool = os.getenv('SOVEREIGN_DEBUG', False)
    sentry_dsn: str = os.getenv('SOVEREIGN_SENTRY_DSN')
    node_match_key: str = os.getenv('SOVEREIGN_NODE_MATCH_KEY', 'cluster')
    node_matching: bool = os.getenv('SOVEREIGN_MATCHING_ENABLED', True)
    source_match_key: str = os.getenv('SOVEREIGN_SOURCE_MATCH_KEY', 'service_clusters')
    sources_refresh_rate: int = os.getenv('SOVEREIGN_SOURCES_REFRESH_RATE', 30)
    refresh_context: bool = os.getenv('SOVEREIGN_REFRESH_CONTEXT', False)
    context_refresh_rate: int = os.getenv('SOVEREIGN_CONTEXT_REFRESH_RATE', 3600)
    dns_hard_fail: bool = os.getenv('SOVEREIGN_DNS_HARD_FAIL', False)

    @property
    def passwords(self):
        return self.auth_passwords.split(',') or []

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        kwargs = [
            f'{k}={v}'
            for k, v in self.show().items()
        ]
        return f'SovereignConfig({kwargs})'

    def show(self):
        safe_items = dict()
        for key, value in self.__dict__.items():
            if key in ['auth_passwords', 'encryption_key', 'passwords', 'sentry_dsn']:
                value = 'redacted'
            safe_items[key] = value
        return safe_items



import os
from pkg_resources import get_distribution, resource_filename
from starlette.templating import Jinja2Templates
from sovereign import config_loader
from sovereign.utils.dictupdate import merge
from sovereign.schemas import SovereignConfig, XdsTemplate
from sovereign.logs import LOG

__versionstr__ = get_distribution('sovereign').version
__version__ = tuple(int(i) for i in __versionstr__.split('.'))

html_templates = Jinja2Templates(resource_filename('sovereign', 'templates'))

XDS_TEMPLATES = dict()
CONFIG_FILE = dict()


try:
    CONFIG_PATHS = os.getenv('SOVEREIGN_CONFIG').split(',')
except AttributeError:
    LOG.error('No configuration specified via environment variable SOVEREIGN_CONFIG')
else:
    for path in CONFIG_PATHS:
        CONFIG_FILE = merge(
            obj_a=CONFIG_FILE,
            obj_b=config_loader.load(path),
            merge_lists=True
        )

    config = SovereignConfig(**CONFIG_FILE)

    for version, templates in config.templates.items():
        XDS_TEMPLATES[version] = dict()
        for _type, path in templates.items():
            XDS_TEMPLATES[version][_type] = XdsTemplate(path=path)

    LOG.msg(
        event='startup',
        env=config.environment,
        config=CONFIG_PATHS,
        context=config.template_context,
        templates=config.templates,
        is_debug=config.debug_enabled
    )



from pydantic import BaseModel, Schema
from fastapi import APIRouter, Body
from starlette.responses import UJSONResponse
from sovereign.middlewares import get_request_id
from sovereign.utils.crypto import encrypt, decrypt, generate_key

router = APIRouter()


class EncryptionRequest(BaseModel):
    data: str = Schema(..., title='Text to be encrypted', min_length=1, max_length=65535)
    key: str = Schema(None, title='Optional Fernet encryption key to use to encrypt', min_length=44, max_length=44)


class DecryptionRequest(BaseModel):
    data: str = Schema(..., title='Text to be decrypted', min_length=1, max_length=65535)
    key: str = Schema(..., title='Fernet encryption key to use to decrypt', min_length=44, max_length=44)


class DecryptableRequest(BaseModel):
    data: str = Schema(..., title='Text to be decrypted', min_length=1, max_length=65535)


@router.post('/decrypt', summary='Decrypt provided encrypted data using a provided key', response_class=UJSONResponse)
async def _decrypt(request: DecryptionRequest = Body(None)):
    return {'result': decrypt(request.data, request.key)}


@router.post('/encrypt', summary='Encrypt provided data using this servers key', response_class=UJSONResponse)
async def _encrypt(request: EncryptionRequest = Body(None)):
    return {'result': encrypt(data=request.data, key=request.key)}


@router.post('/decryptable', summary='Check whether data is decryptable by this server', response_class=UJSONResponse)
async def _decryptable(request: DecryptableRequest = Body(None)):
    try:
        decrypt(request.data)
        ret = {}
        code = 200
    except KeyError as e:
        ret = {
            'error': str(e),
            'request_id': get_request_id()
        }
        code = 500
    return UJSONResponse(ret, status_code=code)


@router.get('/generate_key', summary='Generate a new asymmetric encryption key', response_class=UJSONResponse)
def _generate_key():
    return {'result': generate_key()}



import schedule
from fastapi import Body, BackgroundTasks
from fastapi.routing import APIRouter
from starlette.responses import UJSONResponse
from sovereign import discovery, config
from sovereign.statistics import stats
from sovereign.schemas import DiscoveryRequest, DiscoveryResponse
from sovereign.utils.auth import authenticate

router = APIRouter()


@router.post(
    '/discovery:{xds_type}',
    summary='Envoy Discovery Service Endpoint',
    responses={
        200: {
            'model': DiscoveryResponse,
            'description': 'New resources provided'
        },
        config.no_changes_response_code: {
            'description': 'Resources are up-to-date'
        }
    }
)
async def discovery_response(
        xds_type: discovery.DiscoveryTypes,
        background_tasks: BackgroundTasks,
        discovery_request: DiscoveryRequest = Body(None)
):
    authenticate(discovery_request)
    response: dict = await discovery.response(discovery_request, xds_type.value)
    if response['version_info'] == discovery_request.version_info:
        ret = 'No changes'
        code = config.no_changes_response_code
    elif len(response.get('resources', [])) == 0:
        ret = 'No resources found'
        code = 404
    elif response['version_info'] != discovery_request.version_info:
        ret = response
        code = 200
    else:
        ret = 'Unknown Error'
        code = 500

    metrics_tags = [
        f"client_ip:{discovery_request.node.metadata.get('ipv4', '-')}",
        f"client_version:{discovery_request.envoy_version}",
        f"response_code:{code}",
        f"xds_type:{xds_type.value}"
    ]
    metrics_tags += [f"resource:{resource}" for resource in discovery_request.resource_names]
    stats.increment('discovery.rq_total', tags=metrics_tags)
    background_tasks.add_task(schedule.run_pending)
    return UJSONResponse(content=ret, status_code=code)



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
import zlib
import yaml
from yaml.parser import ParserError
from enum import Enum
from jinja2 import meta
from sovereign import XDS_TEMPLATES, config
from sovereign.statistics import stats
from sovereign.context import template_context
from sovereign.sources import match_node
from sovereign.config_loader import jinja_env
from sovereign.schemas import XdsTemplate, DiscoveryRequest
from sovereign.utils.crypto import disabled_suite

try:
    default_templates = XDS_TEMPLATES['default']
except KeyError:
    raise KeyError(
        'Your configuration should contain default templates. For more details, see '
        'https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/tutorial.html#create-templates '
    )


discovery_types = list(XDS_TEMPLATES['default'].keys())
DiscoveryTypes = Enum('DiscoveryTypes', {t: t for t in discovery_types})


@stats.timed('discovery.version_hash_ms')
def version_hash(*args) -> str:
    """
    Creates a 'version hash' to be used in envoy Discovery Responses.
    """
    data: bytes = repr(args).encode()
    version_info = zlib.adler32(data)
    return str(version_info)


def make_context(request: DiscoveryRequest, debug=config.debug_enabled):
    return {
        'instances': match_node(request),
        'resource_names': request.resources,
        'debug': debug,
        **template_context
    }


async def response(request: DiscoveryRequest, xds, debug=config.debug_enabled):
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds: what type of XDS template to use when rendering
    :param debug: switch to control instance loading / exception raising
    :param context: optional alternative context for generation of templates
    :return: An envoy Discovery Response
    """
    metrics_tags = [
        f'xds_type:{xds}',
        f'partition:{request.node.cluster}'
    ]
    with stats.timed('discovery.total_ms', tags=metrics_tags):
        context = make_context(request, debug)
        if request.node.metadata.get('hide_private_keys'):
            context['crypto'] = disabled_suite

        template: XdsTemplate = XDS_TEMPLATES.get(request.envoy_version, default_templates)[xds]
        template_ast = jinja_env.parse(template.source)
        used_variables = meta.find_undeclared_variables(template_ast)
        unused_variables = [key for key in list(context)
                            if key not in used_variables]

        for key in unused_variables:
            context.pop(key, None)

        config_version = version_hash(
            context,
            template.checksum,
            request.node.cluster,
            request.node.build_version,
            request.node.locality,
        )
        if config_version == request.version_info:
            return {'version_info': config_version}

        with stats.timed('discovery.render_ms', tags=metrics_tags):
            rendered = await template.content.render_async(discovery_request=request, **context)
        return parse_envoy_configuration(rendered, config_version, request, xds, debug)


def parse_envoy_configuration(template_output, version_info, request, request_type, debug=config.debug_enabled):
    try:
        resources = yaml.safe_load(template_output)['resources']
    except ParserError:
        if debug:
            raise
        raise ParserError(
            'Failed to load configuration, there may be '
            'a syntax error in the configured templates. '
            f'xds_type:{request_type} envoy_version:{request.envoy_version}'
        )
    configuration = dict()
    configuration['version_info'] = version_info
    configuration['resources'] = [
        resource
        for resource in resources
        if resource_name(resource) in request.resources
    ]
    return configuration


def resource_name(resource):
    return resource.get('name') or resource['cluster_name']



from collections import namedtuple
from cryptography.fernet import Fernet, InvalidToken
from fastapi.exceptions import HTTPException
from sovereign import config

disabled_suite = namedtuple('DisabledSuite', ['encrypt', 'decrypt'])
disabled_suite.encrypt = lambda x: 'Unavailable (No Secret Key)'
disabled_suite.decrypt = lambda x: 'Unavailable (No Secret Key)'

try:
    _cipher_suite = Fernet(config.encryption_key)
    KEY_AVAILABLE = True
except TypeError:
    KEY_AVAILABLE = False
    _cipher_suite = disabled_suite


def encrypt(data: str, key=None) -> str:
    _local_cipher_suite = _cipher_suite
    if key is not None:
        _local_cipher_suite = Fernet(key)
    try:
        encrypted: bytes = _local_cipher_suite.encrypt(data.encode())
    except (InvalidToken, AttributeError):
        raise HTTPException(status_code=400, detail='Encryption failed')
    return encrypted.decode()


def decrypt(data: str, key=None) -> str:
    _local_cipher_suite = _cipher_suite
    if key is not None:
        _local_cipher_suite = Fernet(key)
    try:
        decrypted: bytes = _local_cipher_suite.decrypt(data.encode())
    except (InvalidToken, AttributeError):
        raise HTTPException(status_code=400, detail='Decryption failed')
    return decrypted.decode()


def generate_key() -> str:
    secret: bytes = Fernet.generate_key()
    return secret.decode()
